/*
Copyright (c) 2007, Dennis M. Sosnoski
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
 * Neither the name of JiBX nor the names of its contributors may be used
   to endorse or promote products derived from this software without specific
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package org.jibx.binding.generator;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.jibx.binding.model.BindingElement;
import org.jibx.binding.model.CollectionElement;
import org.jibx.binding.model.IClass;
import org.jibx.binding.model.IncludeElement;
import org.jibx.binding.model.MappingElement;
import org.jibx.binding.model.NestingElementBase;
import org.jibx.binding.model.StructureElement;
import org.jibx.binding.model.StructureElementBase;
import org.jibx.binding.model.ValueElement;
import org.jibx.runtime.BindingDirectory;
import org.jibx.runtime.IBindingFactory;
import org.jibx.runtime.IMarshallable;
import org.jibx.runtime.IMarshallingContext;
import org.jibx.runtime.JiBXException;
import org.jibx.runtime.QName;
import org.jibx.schema.codegen.ReferenceCountMap;
import org.jibx.util.InsertionOrderedMap;
import org.jibx.util.Types;

/**
 * Binding generator implementation.
 */
public class BindingGenerator
{
    /** Binding generation customizations. */
    private final GlobalCustom m_global;
    
    /** Set of class names to be included. */
    private final Set m_includeSet;
    
    /** Set of class names to be ignored. */
    private final Set m_ignoreSet;
    
    /** Set of class names referenced directly as members of other classes. */
    private final Set m_directSet;
    
    /** Set of class names subclassed by other classes in binding. */
    private final Set m_superSet;
    
    /** Map from fully-qualified class name to mapping details. */
    private final Map m_mappingDetailsMap;
    
    /** Map from namespace URI to binding holder. */
    private final InsertionOrderedMap m_uriBindingMap;
    
    /** Target package for binding code generation. */
    private String m_targetPackage;

    /**
     * Create a generator based on a particular set of customizations.
     * 
     * @param glob
     */
    public BindingGenerator(GlobalCustom glob) {
        m_global = glob;
        m_includeSet = new HashSet();
        m_ignoreSet = new HashSet();
        m_directSet = new HashSet();
        m_superSet = new HashSet();
        m_mappingDetailsMap = new HashMap();
        m_uriBindingMap = new InsertionOrderedMap();
    }
    
    /**
     * Check if a class represents a simple value.
     * TODO: implement ClassCustom hooks for this purpose
     *
     * @param type fully qualified class name
     * @return <code>true</code> if simple value, <code>false</code> if not
     */
    public boolean isValueClass(String type) {
        ClassCustom clas = m_global.forceClassCustomization(type);
        IClass sinfo = clas.getClassInformation().getSuperClass();
        if (sinfo != null && "java.lang.Enum".equals(sinfo.getName())) {
            return true;
        } else {
            return false;
        }
    }
    
    /**
     * Check if a class needs to be included in the binding. This checks all
     * members of the class and if necessary superclasses, returning
     * <code>true</code> if any member is ultimately found with a simple value.
     *
     * @param type fully qualified class name
     * @return <code>true</code> if class to be included in binding,
     * <code>false</code> if it should be skipped
     */
    public boolean checkInclude(String type) {
        if (m_includeSet.contains(type)) {
            return true;
        } else if (m_ignoreSet.contains(type)) {
            return false;
        } else {
            
            // check if any members are to be included
            boolean include = false;
            ClassCustom clas = m_global.forceClassCustomization(type);
            for (Iterator iter = clas.getMembers().iterator(); iter.hasNext();) {
                MemberCustom memb = (MemberCustom)iter.next();
                String mtype = memb.isCollection() ?
                    memb.getItemType() : memb.getWorkingType();
                if (Types.isSimpleValue(mtype) || isValueClass(mtype) ||
                    !m_global.isKnownMapping(mtype) || checkInclude(mtype)) {
                    include = true;
                    break;
                }
            }
            if (!include) {
                
                // force superclass handling if appropriate
                if (clas.getMembers().size() > 0 && clas.isUseSuper()) {
                    String stype = clas.getClassInformation().
                        getSuperClass().getName();
                    if (!"java.lang.Object".equals(stype) &&
                        checkInclude(stype)) {
                        include = true;
                    }
                }
            }
            if (include) {
                m_includeSet.add(type);
            } else {
                m_ignoreSet.add(type);
            }
            return include;
        }
    }
    
    /**
     * Expand all references from a class. This checks the types of all members
     * of the class, counting references and calling itself recursively for any
     * types which are not primitives and not java.* or javax.* classes. It also
     * expands the superclass, if specified by the class customizations.
     *
     * @param type fully qualified class name
     * @param refmap reference count map
     */
    public void expandReferences(String type, ReferenceCountMap refmap) {
        if (checkInclude(type)) {
            
            // expand all references from members
            ClassCustom clas = m_global.forceClassCustomization(type);
            for (Iterator iter = clas.getMembers().iterator();
                iter.hasNext();) {
                MemberCustom memb = (MemberCustom)iter.next();
                String mtype = memb.isCollection() ?
                    memb.getItemType() : memb.getWorkingType();
                if (!Types.isSimpleValue(mtype) && !isValueClass(mtype) &&
                    !m_global.isKnownMapping(mtype)) {
                    if ((mtype.startsWith("java.")) ||
                        (mtype.startsWith("javax."))) {
                        throw new IllegalStateException("No way to handle type '" +
                            mtype + '\'');
                    } else if (checkInclude(mtype)) {
                        if (refmap.incrementCount(mtype) <= 1) {
                            expandReferences(mtype, refmap);
                        }
                        m_directSet.add(mtype);
                    }
                }
            }
            
            // force superclass handling if appropriate
            if (clas.getMembers().size() > 0 && clas.isUseSuper()) {
                String stype = clas.getClassInformation().
                    getSuperClass().getName();
                if (!"java.lang.Object".equals(stype) &&
                    !Types.isSimpleValue(stype) && !isValueClass(stype) &&
                    !m_global.isKnownMapping(stype) && checkInclude(stype)) {
                    if (refmap.incrementCount(stype) <= 1) {
                        expandReferences(stype, refmap);
                    }
                    m_superSet.add(stype);
                }
            }
        }
    }
    
    /**
     * Set creation information for structure binding component. This includes
     * the declared type, as well as the create type and factory method.
     *
     * @param memb
     * @param struct
     */
    private void setTypes(MemberCustom memb, StructureElementBase struct) {
        String type = memb.getActualType();
        if (type != null && !type.equals(memb.getStatedType())) {
            struct.setDeclaredType(type);
        }
        struct.setCreateType(memb.getCreateType());
        struct.setFactoryName(memb.getFactoryMethod());
    }
    
    /**
     * General object comparison method. Don't know why Sun hasn't seen fit to
     * include this somewhere, but at least it's easy to write (over and over
     * again).
     * 
     * @param a first object to be compared
     * @param b second object to be compared
     * @return <code>true</code> if both objects are <code>null</code>, or if
     * <code>a.equals(b)</code>; <code>false</code> otherwise
     */
    public static boolean isEqual(Object a, Object b) {
        return (a == null) ? b == null : a.equals(b);
    }

    /**
     * Check for dependency on another binding.
     *
     * @param uri namespace for binding of referenced component
     * @param hold binding holder
     */
    private void checkDependency(String uri, BindingHolder hold) {
        if (!isEqual(uri, hold.getNamespace())) {
            BindingHolder tohold = findBinding(uri);
            hold.addReference(tohold);
        }
    }

    /**
     * Define the details of a collection binding. Collection bindings may be
     * empty (in the case where the item type is directly mapped), have a
     * &lt;value> child element, or have either a mapping reference or direct
     * definition &lt;structure> child element. This determines the appropriate
     * form based on the item type.
     *
     * @param itype
     * @param iname
     * @param coll
     * @param hold
     */
    public void defineCollection(String itype, String iname,
        CollectionElement coll, BindingHolder hold) {
        
        // check item type handling
        MappingDetail detail = (MappingDetail)m_mappingDetailsMap.get(itype);
        if (detail != null) {
            ClassCustom icust = m_global.getClassCustomization(itype);
            if (detail.isUseAbstract() && !detail.isExtended()) {
                
                // add reference to appropriate binding
                QName qname = detail.getTypeQName();
                hold.addReference(findBinding(qname.getUri()));
                
                // add <structure> with map-as for abstract mapping
                if (iname == null) {
                    iname = icust.getElementName();
                }
                StructureElement struct = new StructureElement();
                if (qname == null) {
                    struct.setMapAsName(itype);
                } else {
                    
                    // set the type reference
                    struct.setMapAsQName(qname);
                    
                    // check dependency on other binding
                    checkDependency(qname.getUri(), hold);
                    
                }
                struct.setName(iname);
                struct.setCreateType(icust.getCreateType());
                struct.setFactoryName(icust.getFactoryMethod());
                coll.addChild(struct);
                
            } else {
                
                // just set item-type for concrete mapping
                coll.setItemTypeName(itype);
                
            }
            
        } else if (m_global.isKnownMapping(itype)) {
            
            // just set item-type for known mapping
            coll.setItemTypeName(itype);
            
        } else if (Types.isSimpleValue(itype) || isValueClass(itype)) {
            
            // add <value> for simple type
            ValueElement value = new ValueElement();
            value.setName(iname);
            value.setDeclaredType(itype);
            coll.addChild(value);
            
        } else {
            
            // embed structure definition within the collection
            StructureElement struct = new StructureElement();
            struct.setDeclaredType(itype);
            ClassCustom icust = m_global.getClassCustomization(itype);
            struct.setCreateType(icust.getCreateType());
            struct.setFactoryName(icust.getFactoryMethod());
            if (iname == null) {
                iname = icust.getElementName();
            }
            struct.setName(iname);
            addMemberBindings(icust, struct, hold);
            coll.addChild(struct);
            
        }
    }
    
    /**
     * Add binding details for all members of a class.
     *
     * @param cust class customization information
     * @param parent containing binding component
     * @param hold binding holder
     */
    private void addMemberBindings(ClassCustom cust,
        NestingElementBase parent, BindingHolder hold) {
        
        // generate binding component for each member of class
        for (Iterator iter = cust.getMembers().iterator(); iter.hasNext();) {
            MemberCustom memb = (MemberCustom)iter.next();
            if (memb.isCollection()) {
                
                // create collection element for field or property
                CollectionElement coll = new CollectionElement();
                if (memb.getFieldName() == null) {
                    coll.setGetName(memb.getGetName());
                    coll.setSetName(memb.getSetName());
                } else {
                    coll.setFieldName(memb.getFieldName());
                }
                setTypes(memb, coll);
                
                // set the element name, if defined
                String name = memb.getXmlName();
                if (name != null) {
                    coll.setName(name);
                }
                
                // check for optional value
                if (!memb.isRequired()) {
                    coll.setUsageName("optional");
                }
                
                // check for type defined
                String itype = memb.getItemType();
                if (itype != null) {
                    
                    // set name to be used for items in collection
                    String iname = memb.getItemName();
                    if (iname == null) {
                        
                        // check for collection name that looks like a plural
                        if (name != null && name.endsWith("s")) {
                            
                            // convert plural ending to singular
                            if (name.endsWith("ies")) {
                                iname = name.substring(0, name.length()-3) + "y";
                            } else {
                                iname = name.substring(0, name.length()-1);
                            }
                            
                        } else {
                            
                            // use name based on item type
                            String simple = itype;
                            int split = simple.lastIndexOf('.');
                            if (split >= 0) {
                                simple = simple.substring(0, split);
                            }
                            iname = cust.convertName(simple);
                            
                        }
                    }
                    
                    // define the collection details
                    defineCollection(itype, iname, coll, hold);
                    
                }
                parent.addChild(coll);
                
            } else {
                
                // check whether value or structure
                String wtype = memb.getWorkingType();
                if (Types.isSimpleValue(wtype) || isValueClass(wtype)) {
                    
                    // add <value> for simple type
                    ValueElement value = new ValueElement();
                    if (memb.getFieldName() == null) {
                        value.setGetName(memb.getGetName());
                        value.setSetName(memb.getSetName());
                    } else {
                        value.setFieldName(memb.getFieldName());
                    }
                    value.setName(memb.getXmlName());
                    
                    // check for optional value
                    if (!memb.isRequired()) {
                        value.setUsageName("optional");
                    }
                    
                    // configure added property values
                    int style = memb.getStyle();
                    if (style == NestingBase.ATTRIBUTE_VALUE_STYLE) {
                        value.setStyleName("attribute");
                    } else if (style == NestingBase.ELEMENT_VALUE_STYLE) {
                        value.setStyleName("element");
                    } else {
                        value.setStyleName("text");
                    }
                    if (memb.getActualType() != null) {
                        value.setDeclaredType(memb.getActualType());
                    }
                    parent.addChild(value);
                    
                } else {
                    
                    // create <structure> with name and access information
                    StructureElement struct = new StructureElement();
                    if (memb.getFieldName() == null) {
                        struct.setGetName(memb.getGetName());
                        struct.setSetName(memb.getSetName());
                    } else {
                        struct.setFieldName(memb.getFieldName());
                    }
                    setTypes(memb, struct);
                    
                    // check for optional value
                    if (!memb.isRequired()) {
                        struct.setUsageName("optional");
                    }
                    
                    // set details based on class handling
                    MappingDetail detail =
                        (MappingDetail)m_mappingDetailsMap.get(wtype);
                    if (detail != null) {
                        if (detail.isUseAbstract() && !detail.isExtended()) {
                            
                            // add 'map-as' and 'name' for abstract mapping
                            QName qname = detail.getTypeQName();
                            if (qname == null) {
                                struct.setMapAsName(wtype);
                            } else {
                                
                                // set type name reference from structure
                                struct.setMapAsQName(qname);
                                
                                // create binding dependency, if needed
                                checkDependency(qname.getUri(), hold);
                            }
                            struct.setName(memb.getXmlName());
                            
                        } else {
                            
                            // create binding dependency, if needed
                            checkDependency(detail.getElementQName().getUri(),
                                hold);
                            
                        }
                    } else if (!m_global.isKnownMapping(wtype)){
                        
                        // check for type needed
                        if (memb.getActualType() != null) {
                            struct.setDeclaredType(memb.getActualType());
                        }
                        if (memb.getCreateType() != null) {
                            struct.setCreateType(memb.getCreateType());
                        }
                        if (memb.getFactoryMethod() != null) {
                            struct.setFactoryName(memb.getFactoryMethod());
                        }
                        
                        // add a name if an optional value
                        if (!memb.isRequired()) {
                            struct.setName(memb.getXmlName());
                        }
                        
                        // handle inline structure definition
                        ClassCustom icust =
                            m_global.getClassCustomization(memb.getWorkingType());
                        addMemberBindings(icust, struct, hold);
                        
                    }
                    parent.addChild(struct);
                    
                }
            }
        }
    }
    
    /**
     * Add the &lt;mapping> definition for a class to a binding. This creates
     * either an abstract mapping with a type name, or a concrete mapping with
     * an element name, as determined by the passed-in mapping information. If
     * the class is a concrete mapping that extends or implements another class
     * with an anonymous abstract mapping, the created &lt;mapping> will extend
     * that base mapping.
     * TODO: type substitution requires extending the binding definition
     *
     * @param type fully qualified class name
     * @param detail mapping details
     */
    private void addMapping(String type, MappingDetail detail) {
        
        // create the basic mapping structure
        ClassCustom cust = m_global.forceClassCustomization(type);
        MappingElement mapping = new MappingElement();
        mapping.setClassName(type);
        mapping.setCreateType(cust.getCreateType());
        mapping.setFactoryName(cust.getFactoryMethod());
        
        // check type of definition required
        BindingHolder hold;
        if (detail.isUseAbstract()) {
            
            // create abstract mapping
            mapping.setAbstract(true);
            QName qname = detail.getTypeQName();
            mapping.setTypeQName(qname);
            hold = findBinding(qname.getUri());
            
        } else {
            
            // concrete mapping, just need simple name (namespace from binding)
            QName qname = detail.getElementQName();
            mapping.setName(qname.getName());
            hold = findBinding(qname.getUri());
            
        }
        
        // check for superclass mapping to be extended (do this first for
        //  compatibility with schema type extension)
        IClass clas = cust.getClassInformation();
        StructureElement struct = null;
        String ptype = detail.getExtendsType();
        if (ptype != null) {
            
            // create reference to parent mapping
            struct = new StructureElement();
            MappingDetail pdetail =
                (MappingDetail)m_mappingDetailsMap.get(ptype);
            if (pdetail.isUseAbstract()) {
                
                // reference abstract mapped superclass
                QName qname = pdetail.getTypeQName();
                if (qname == null) {
                    throw new IllegalStateException("Internal error: unimplemented case");
                    // anonymous 
//                    struct.setMapAsName(ptype);
                } else {
                    
                    // set extension base type and reference
                    mapping.setExtendsName(ptype);
                    struct.setMapAsQName(qname);
                    
                    // create binding dependency, if needed
                    checkDependency(qname.getUri(), hold);
                }
                
            } else {
                
                // reference concrete mapped superclass
                struct.setMapAsName(ptype);
                mapping.setExtendsName(ptype);
                
            }
            
        } else {
            
            // not extending a base mapping, check if need to include superclass
            IClass parent = clas.getSuperClass();
            if (cust.isUseSuper() && parent != null) {
                ptype = parent.getName();
                ClassCustom scust = m_global.getClassCustomization(ptype);
                if (scust != null && scust.getMembers().size() > 0) {
                    MappingDetail pdetail =
                        (MappingDetail)m_mappingDetailsMap.get(ptype);
                    if (pdetail == null) {
                        
                        // handle parent details inline, if needed
                        struct = new StructureElement();
                        struct.setDeclaredType(ptype);
                        addMemberBindings(scust, struct, hold);
                        if (struct.children().size() == 0) {
                            struct = null;
                        }
                        
                    } else {
                        throw new IllegalStateException("Internal error: unsupported combination");
                    }
                }
                if (struct == null) {
                    
                    // check for interface with abstract mapping to extend
                    if (!detail.isUseAbstract() &&
                        mapping.getExtendsName() == null) {
                        String[] intfs = parent.getInterfaces();
                        for (int i = 0; i < intfs.length; i++) {
                            String intf = intfs[i];
                            MappingDetail idetail =
                                (MappingDetail)m_mappingDetailsMap.get(intf);
                            if (idetail != null && idetail.isUseAbstract() &&
                                idetail.getTypeQName() == null) {
                                
                                // reference abstract mapped interface
                                struct = new StructureElement();
                                struct.setMapAsName(intf);
                                mapping.setExtendsName(intf);
                                break;
                            }
                        }
                    }
                    
                }
            }
        }
        if (struct != null) {
            mapping.addChild(struct);
        }
        
        // add all details of class member handling to binding
        addMemberBindings(cust, mapping, hold);
        hold.getBinding().addTopChild(mapping);
        if (mapping.isAbstract()) {
            
            // check for both abstract and concrete mapping needed
            detail.setAbstractMapping(mapping);
            if (detail.isUseConcrete()) {
                
                // reference abstract mapping from concrete mapping
                mapping = new MappingElement();
                mapping.setClassName(type);
                mapping.setCreateType(cust.getCreateType());
                mapping.setFactoryName(cust.getFactoryMethod());
                mapping.setName(cust.getElementName());
                
                // set create type for concrete mapping of abstract class
                if (clas.isAbstract()) {
                    mapping.setAbstract(true);
                }
                
                // define content as structure reference to abstract mapping
                struct = new StructureElement();
                QName qname = detail.getTypeQName();
                struct.setMapAsQName(qname);
                mapping.addChild(struct);
                hold = findBinding(qname.getUri());
                hold.getBinding().addTopChild(mapping);
                detail.setConcreteMapping(mapping);
            }
            
        } else {
            detail.setConcreteMapping(mapping);
        }
    }

    /**
     * Find the binding to be used for a particular namespace. If this is the
     * first time a particular namespace was requested, a new binding will be
     * created for that namespace and returned.
     *
     * @param uri namespace URI (<code>null</code> if no namespace)
     * @return binding holder
     */
    public BindingHolder findBinding(String uri) {
        BindingHolder hold = (BindingHolder)m_uriBindingMap.get(uri);
        if (hold == null) {
            hold = new BindingHolder(uri, m_uriBindingMap.size()+1);
            m_uriBindingMap.put(uri, hold);
            BindingElement binding = hold.getBinding();
            binding.setForceClasses(m_global.isForceClasses());
            binding.setTrackSource(m_global.isTrackSource());
            binding.setAddConstructors(m_global.isAddConstructors());
            binding.setInBinding(m_global.isInput());
            binding.setOutBinding(m_global.isOutput());
        }
        return hold;
    }

    /**
     * Add the details for mapping a class.
     *
     * @param abstr force abstract mapping flag
     * @param type fully-qualified class name
     * @return mapping details
     */
    private MappingDetail addMappingDetails(Boolean abstr, String type) {
        
        // check for existing detail
        MappingDetail detail = (MappingDetail)m_mappingDetailsMap.get(type);
        if (detail == null) {
            
            // check if superclass is base for extension
            ClassCustom cust = m_global.getClassCustomization(type);
            String stype = null;
            IClass sclas = cust.getClassInformation().getSuperClass();
            while (sclas != null) {
                if (m_directSet.contains(sclas.getName())) {
                    stype = sclas.getName();
                    Boolean isabs = Boolean.valueOf(sclas.isAbstract());
                    MappingDetail sdetail = addMappingDetails(isabs, stype);
                    sdetail.setExtended(true);
                    break;
                } else {
                    sclas = sclas.getSuperClass();
                }
            }
            
            // create mapping details as specified
            boolean abs = (abstr == null) ?
                cust.isMapAbstract() : abstr.booleanValue();
            detail = new MappingDetail(cust.getTypeQName(),
                cust.getElementQName(), stype);
            if (abs || !cust.isConcrete()) {
                detail.setUseAbstract(true);
            } else {
                detail.setUseConcrete(true);
            }
            m_mappingDetailsMap.put(type, detail);
            
            // force concrete mapping if extension or base for extension
            if (abs && (sclas != null || m_superSet.contains(type))) {
                detail.setUseConcrete(true);
            }
            
            // check if package usable as target for binding code
            if (m_targetPackage == null &&
                cust.getClassInformation().isModifiable()) {
                m_targetPackage = ((PackageCustom)cust.getParent()).getName();
            }
            
        } else if (abstr != null) {
            
            // mapping detail already generated, just make sure of variety
            if (abstr.booleanValue()) {
                detail.setUseAbstract(true);
            } else {
                detail.setUseConcrete(true);
            }
            
        }
        return detail;
    }

    /**
     * Find closure of references from a supplied list of classes. References
     * counted, and direct references are accumulated for handling. The supplied
     * list may include generic classes with type parameters.
     *
     * @param classes
     * @param refmap
     */
    private void findReferences(List classes, ReferenceCountMap refmap) {
        for (int i = 0; i < classes.size(); i++) {
            String type = (String)classes.get(i);
            int split = type.indexOf('<');
            if (split > 0) {
                type = type.substring(split+1, type.length()-1);
            }
            expandReferences(type, refmap);
        }
    }

    /**
     * Add mapping details for classes referenced more than once.
     *
     * @param refmap
     */
    private void addReferencedMappings(ReferenceCountMap refmap) {
        for (Iterator iter = refmap.iterator(); iter.hasNext();) {
            String type = (String)iter.next();
            if (refmap.getCount(type) > 1 &&
                !m_mappingDetailsMap.containsKey(type)) {
                addMappingDetails(null, type);
            }
        }
    }

    /**
     * Fix all references between bindings. This needs to be called after the
     * generation of binding components is completed, in order to make sure that
     * required namespace definitions are included in the output bindings.
     */
    private void fixBindingReferences() {
        ArrayList uris = getNamespaces();
        for (int i = 0; i < uris.size(); i++) {
            String uri = (String)uris.get(i);
            BindingHolder holder = (BindingHolder)m_uriBindingMap.get(uri);
            holder.fixReferences();
        }
    }

    /**
     * Generate the mapping definitions for classes referenced more than once.
     *
     * @param refmap
     */
    private void generateReferencedMappings(ReferenceCountMap refmap) {
        for (Iterator iter = refmap.iterator(); iter.hasNext();) {
            String type = (String)iter.next();
            if (refmap.getCount(type) > 1) {
                MappingDetail detail =
                    (MappingDetail)m_mappingDetailsMap.get(type);
                if (!detail.isGenerated()) {
                    addMapping(type, detail);
                    detail.setGenerated(true);
                }
            }
        }
    }

    /**
     * Generate mappings for a list of classes. The mapping details must have
     * been configured before this method is called.
     *
     * @param classes
     */
    private void generateMappings(List classes) {
        for (int i = 0; i < classes.size(); i++) {
            String type = (String)classes.get(i);
            MappingDetail detail = (MappingDetail)m_mappingDetailsMap.get(type);
            if (!detail.isGenerated()) {
                addMapping(type, detail);
                detail.setGenerated(true);
            }
        }
    }
    
    /**
     * Fix the base classes that are to be used as extension types. This step is
     * needed to generate the substitution group structures for classes which
     * are both referenced directly and extended by other classes. The method
     * must be called after all mapping details have been constucted but before
     * the actual binding generation.
     */
    private void fixBaseClasses() {
        for (Iterator iter = m_directSet.iterator(); iter.hasNext();) {
            String type = (String)iter.next();
            MappingDetail detail = (MappingDetail)m_mappingDetailsMap.get(type);
            if (detail != null && detail.isExtended()) {
                detail.setUseConcrete(true);
            }
        }
    }
    
    /**
     * Generate binding(s) for a list of classes. This creates a &lt;mapping>
     * definition for each class in the list, and either embeds &lt;structure>
     * definitions or creates separate &lt;mapping>s for other classes
     * referenced by these classes. If all the classes use the same namespace
     * only the binding for that namespace will be created; otherwise, a
     * separate binding will be created for each namespace.
     *
     * @param abstr force abstract mapping flag (determined by class
     * customizations if <code>null</code>)
     * @param classes class list
     */
    public void generate(Boolean abstr, List classes) {
        
        // start by expanding and counting references from supplied classes
        ReferenceCountMap refmap = new ReferenceCountMap();
        findReferences(classes, refmap);
        
        // set the classes to be handled with <mapping> definitions
        for (int i = 0; i < classes.size(); i++) {
            addMappingDetails(abstr, (String)classes.get(i));
        }
        addReferencedMappings(refmap);
        fixBaseClasses();
        
        // generate the binding(s)
        generateMappings(classes);
        generateReferencedMappings(refmap);
        fixBindingReferences();
    }
    
    /**
     * Generate binding(s) for lists of classes. This creates a &lt;mapping>
     * definition for each class in the lists, and either embeds &lt;structure>
     * definitions or creates separate &lt;mapping>s for other classes
     * referenced by these classes. If all the classes use the same namespace
     * only the binding for that namespace will be created; otherwise, a
     * separate binding will be created for each namespace.
     * 
     * @param qnames list of names for concrete mappings
     * @param concrs list of classes to be given concrete mappings
     * @param abstrs list of classes to be given abstract mappings
     */
    public void generateSpecified(ArrayList qnames, List concrs,
        List abstrs) {
        
        // start by expanding and counting references from supplied classes
        ReferenceCountMap refmap = new ReferenceCountMap();
        findReferences(concrs, refmap);
        findReferences(abstrs, refmap);
        
        // set classes to be handled with mapping definitions
        for (int i = 0; i < concrs.size(); i++) {
            MappingDetail detail =
                addMappingDetails(Boolean.FALSE, (String)concrs.get(i));
            detail.setElementQName((QName)qnames.get(i));
        }
        for (int i = 0; i < abstrs.size(); i++) {
            addMappingDetails(Boolean.TRUE, (String)abstrs.get(i));
        }
        addReferencedMappings(refmap);
        fixBaseClasses();
        
        // generate the binding(s)
        generateMappings(concrs);
        generateMappings(abstrs);
        generateReferencedMappings(refmap);
        fixBindingReferences();
    }
    
    /**
     * Get the mapping details for a class. This method should only be used
     * after the {@link #generate(boolean, ArrayList)} method has
     * been called.
     *
     * @param type fully-qualified class name
     * @return mapping details, or <code>null</code> if none
     */
    public MappingDetail getMappingDetail(String type) {
        return (MappingDetail)m_mappingDetailsMap.get(type);
    }
    
    /**
     * Get the binding definition for a namespace. This method should only be
     * used after the {@link #generate(boolean, ArrayList)} method
     * has been called.
     *
     * @param uri
     * @return binding holder, or <code>null</code> if none
     */
    public BindingHolder getBinding(String uri) {
        return (BindingHolder)m_uriBindingMap.get(uri);
    }

    /**
     * Get the list of binding namespace URIs.
     *
     * @return namespaces
     */
    public ArrayList getNamespaces() {
        return m_uriBindingMap.keyList();
    }
    
    /**
     * Check if a character is an ASCII alpha character.
     * 
     * @param chr
     * @return alpha character flag
     */
    private static boolean isAsciiAlpha(char chr) {
        return (chr >= 'a' && chr <= 'z') || (chr >= 'A' && chr <= 'Z');
    }
    
    /**
     * Check if a character is an ASCII numeric character.
     * 
     * @param chr
     * @return numeric character flag
     */
    private static boolean isAsciiNum(char chr) {
        return chr >= '0' && chr <= '9';
    }
    
    /**
     * Check if a character is an ASCII alpha or numeric character.
     * 
     * @param chr
     * @return alpha or numeric character flag
     */
    private static boolean isAsciiAlphaNum(char chr) {
        return isAsciiAlpha(chr) || isAsciiNum(chr);
    }

    /**
     * Configure the names to be used for writing bindings to files. If only one
     * binding has been defined, it just gets the supplied name. If multiple
     * bindings have been defined, a single root binding is constructed which
     * includes all the other bindings, and that root binding is given the
     * supplied names while the other bindings are given unique names within the
     * same directory.
     *
     * @param name file name for root or singleton binding definition
     * @return root or singleton binding holder
     */
    public BindingHolder configureBindings(String name) {
        BindingHolder rhold;
        BindingElement root;
        ArrayList uris = getNamespaces();
        if (uris.size() == 1) {
            
            // single binding, just write it using supplied name
            rhold = (BindingHolder)m_uriBindingMap.get(uris.get(0));
            root = rhold.getBinding();
            
        } else {
            
            // get or create no namespace binding
            rhold = findBinding(null);
            root = rhold.getBinding();
            
            // set file names and add to root binding
            Set nameset = new HashSet();
            for (int i = 0; i < uris.size(); i++) {
                String uri = (String)uris.get(i);
                BindingHolder holder = (BindingHolder)m_uriBindingMap.get(uri);
                if (holder != rhold) {
                    
                    // get last part of namespace URI as file name candidate
                    String bindname;
                    String raw = holder.getNamespace();
                    if (raw == null) {
                        bindname = "nonamespaceBinding";
                    } else {
                        int split = raw.lastIndexOf(':');
                        raw = raw.substring(split+1);
                        split = raw.lastIndexOf('/');
                        raw = raw.substring(split+1);
                        split = raw.lastIndexOf('\\');
                        raw = raw.substring(split+1);
                        
                        // eliminate any invalid characters in name
                        StringBuffer buff = new StringBuffer();
                        int index = 0;
                        char chr = raw.charAt(0);
                        if (isAsciiAlpha(chr)) {
                            buff.append(chr);
                            index = 1;
                        } else {
                            buff.append('_');
                        }
                        while (index < raw.length()) {
                            chr = raw.charAt(index++);
                            if (isAsciiAlphaNum(chr)) {
                                buff.append(chr);
                            }
                        }
                        buff.append("Binding");
                        bindname = buff.toString();
                    }
                    
                    // ensure uniqueness of the name
                    String uname = bindname.toLowerCase();
                    int pass = 0;
                    while (nameset.contains(uname)) {
                        bindname = bindname + pass;
                        uname = bindname.toLowerCase();
                    }
                    nameset.add(uname);
                    holder.setFileName(bindname + ".xml");
                    
                    // include within the root binding
                    IncludeElement include = new IncludeElement();
                    include.setIncludePath(holder.getFileName());
                    root.addTopChild(include);
                    
                }
            }
        }
        
        // set the file name on the singleton or root binding
        rhold.setFileName(name);
        
        // set the binding name based on the file name
        int split = name.lastIndexOf('.');
        if (split > 0) {
            root.setName(name.substring(0, split));
        } else {
            root.setName(name);
        }
        
        // set the target package for code generation
        root.setTargetPackage(m_targetPackage);
        return rhold;
    }
    
    /**
     * Write the bindings to supplied destination path.
     *
     * @param dir target directory for writing binding definitions, output to
     * console if <code>null</code>
     * @throws JiBXException
     * @throws IOException
     */
    public void writeBindings(File dir) throws JiBXException, IOException {
        
        // write the bindings
        IBindingFactory fact =
            BindingDirectory.getFactory(BindingElement.class);
        IMarshallingContext ictx = fact.createMarshallingContext();
        ictx.setIndent(2);
        ArrayList uris = getNamespaces();
        for (int i = 0; i < uris.size(); i++) {
            String uri = (String)uris.get(i);
            BindingHolder holder = (BindingHolder)m_uriBindingMap.get(uri);
            if (dir == null) {
                ictx.setOutput(System.out, null);
            } else {
                File file = new File(dir, holder.getFileName());
                ictx.setOutput(new FileOutputStream(file), null);
            }
            ((IMarshallable)holder.getBinding()).marshal(ictx);
            ictx.getXmlWriter().flush();
        }
    }

    /**
     * Run the binding generation using command line parameters.
     *
     * @param args
     * @throws JiBXException
     * @throws IOException
     */
    public static void main(String[] args) throws JiBXException, IOException {
        BindingGeneratorCommandLine parms = new BindingGeneratorCommandLine();
        if (args.length > 0 && parms.processArgs(args)) {
            
            // generate bindings for all namespaces
            BindingGenerator gen = new BindingGenerator(parms.getGlobal());
            gen.generate(parms.getAbstract(), parms.getExtraArgs());
            
            // write the bindings
            gen.configureBindings(parms.getBindingName());
            gen.writeBindings(parms.getGeneratePath());
            
        } else {
            if (args.length > 0) {
                System.err.println("Terminating due to command line errors");
            } else {
                parms.printUsage();
            }
            System.exit(1);
        }
    }
    
    /**
     * Holder for the details of how a class is going to be mapped. This
     * information is needed in order to determine how to reference the mapped
     * class when it's used within another mapping definition.
     */
    public static class MappingDetail
    {
        /** Generate abstract mapping flag. */
        private boolean m_useAbstract;
        
        /** Generate concrete mapping flag. */
        private boolean m_useConcrete;
        
        /** Flag for class extended by other mapped class(es). */
        private boolean m_isExtended;
        
        /** Abstract mapping type name (<code>null</code> if none). */
        private final QName m_typeQName;
        
        /** Concrete mapping element name (<code>null</code> if none). */
        private QName m_elementQName;
        
        /** Concrete mapping type extended by this one. */
        private String m_extendsType;
        
        /** Flag for mapping(s) has been generated. */
        private boolean m_isGenerated;
        
        /** Abstract mapping definition (<code>null</code> if none). */
        private MappingElement m_abstractMapping;
        
        /** Concrete mapping definition (<code>null</code> if none). */
        private MappingElement m_concreteMapping;
        
        /**
         * Constructor.
         * 
         * @param aname abstract mapping type name
         * @param cname concrete mapping element name
         * @param stype superclass for extension (<code>null</code> if not an
         * extension mapping)
         */
        protected MappingDetail(QName aname, QName cname, String stype) {
            m_typeQName = aname;
            m_elementQName = cname;
            m_extendsType = stype;
        }

        public MappingElement getAbstractMapping() {
            return m_abstractMapping;
        }

        public void setAbstractMapping(MappingElement abstractMapping) {
            m_abstractMapping = abstractMapping;
        }

        public MappingElement getConcreteMapping() {
            return m_concreteMapping;
        }

        public void setConcreteMapping(MappingElement concreteMapping) {
            m_concreteMapping = concreteMapping;
        }

        public boolean isExtended() {
            return m_isExtended;
        }

        public void setExtended(boolean isExtended) {
            m_isExtended = isExtended;
        }

        public boolean isGenerated() {
            return m_isGenerated;
        }

        public void setGenerated(boolean isGenerated) {
            m_isGenerated = isGenerated;
        }

        public boolean isUseAbstract() {
            return m_useAbstract;
        }

        public void setUseAbstract(boolean useAbstract) {
            m_useAbstract = useAbstract;
        }

        public boolean isUseConcrete() {
            return m_useConcrete;
        }

        public void setUseConcrete(boolean useConcrete) {
            m_useConcrete = useConcrete;
        }

        public QName getElementQName() {
            return m_elementQName;
        }

        public void setElementQName(QName elementQName) {
            m_elementQName = elementQName;
        }

        public String getExtendsType() {
            return m_extendsType;
        }

        public QName getTypeQName() {
            return m_typeQName;
        }
    }
}